from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QGroupBox, QTableView, QSpinBox, QAction, QToolBar, \
    QHeaderView, QComboBox, QDoubleSpinBox, QWidget, QSizePolicy, QSlider, QWidgetAction, QFrame, QLineEdit

base_view = ["Device"]
default_views = {
    "Home": base_view + ["Module", "Power", "Color", "LoadAvg", "LinkCount", "Uptime"],
    "Health": base_view + ["Uptime", "BootCount", "RestartReason", "LoadAvg", "Sleep", "MqttCount", "LinkCount", "Downtime", "RSSI"],
    "Firmware": base_view + ["Version", "Core", "SDK", "ProgramSize", "Free", "OtaUrl"],
    "Wifi": base_view + ["Hostname", "Mac", "IPAddress", "Gateway", "SSId", "BSSId", "Channel", "RSSI", "LinkCount", "Downtime"],
    "MQTT": base_view + ["Topic", "FullTopic", "CommandTopic", "StatTopic", "TeleTopic", "FallbackTopic", "GroupTopic"],
}

console_font = QFont("asd")
console_font.setStyleHint(QFont.TypeWriter)

docs_url = "https://tasmota.github.io/docs/"


class VLayout(QVBoxLayout):
    def __init__(self, margin=3, spacing=3, label='', *args, **kwargs):
        super(VLayout, self).__init__(*args, **kwargs)
        if isinstance(margin, int):
            self.setContentsMargins(margin, margin, margin, margin)
        elif isinstance(margin, list):
            self.setContentsMargins(margin[0], margin[1], margin[2], margin[3])

        self.setSpacing(spacing)

        if label:
            self.addWidget(QLabel(label))

    def addWidgets(self, widgets):
        for w in widgets:
            self.addWidget(w)

    def addSpacer(self):
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.addWidget(spacer)


class HLayout(QHBoxLayout):
    def __init__(self, margin=3, spacing=3, label='', *args, **kwargs):
        super(HLayout, self).__init__(*args, **kwargs)
        if isinstance(margin, int):
            self.setContentsMargins(margin, margin, margin, margin)
        elif isinstance(margin, list):
            self.setContentsMargins(margin[0], margin[1], margin[2], margin[3])
        self.setSpacing(spacing)

        if label:
            self.addWidget(QLabel(label))

    def addWidgets(self, widgets):
        for w in widgets:
            self.addWidget(w)

    def addSpacer(self):
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.addWidget(spacer)


class GroupBoxV(QGroupBox):
    def __init__(self, title, margin=3, spacing=3, *args, **kwargs):
        super(GroupBoxV, self).__init__(*args, **kwargs)

        self.setTitle(title)

        layout = VLayout()
        layout.setSpacing(spacing)

        if isinstance(margin, int):
            layout.setContentsMargins(margin, margin, margin, margin)
        elif isinstance(margin, list):
            layout.setContentsMargins(margin[0], margin[1], margin[2], margin[3])

        self.setLayout(layout)

    def addWidget(self, w):
        self.layout().addWidget(w)

    def addWidgets(self, widgets):
        for w in widgets:
            self.layout().addWidget(w)

    def addLayout(self, w):
        self.layout().addLayout(w)


class GroupBoxH(QGroupBox):
    def __init__(self, title, margin=None, spacing=None, *args, **kwargs):
        super(GroupBoxH, self).__init__(title)
        self.setLayout(HLayout())

    def addWidget(self, w):
        self.layout().addWidget(w)

    def addWidgets(self, widgets):
        for w in widgets:
            self.layout().addWidget(w)

    def addLayout(self, w):
        self.layout().addLayout(w)


class TableView(QTableView):
    def __init__(self, *args, **kwargs):
        super(TableView, self).__init__(*args, **kwargs)
        self.setAlternatingRowColors(True)

        self.horizontalHeader().setHighlightSections(False)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.verticalHeader().setVisible(False)
        self.verticalHeader().setHighlightSections(False)
        # self.verticalHeader().setDefaultSectionSize(24)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        self.setSelectionBehavior(QTableView.SelectRows)
        self.setSelectionMode(QTableView.SingleSelection)

        self.setEditTriggers(self.NoEditTriggers)

    def setColumnTitles(self, titles):
        for col, title in enumerate(titles):
            self.model().setHeaderData(col, Qt.Horizontal, title)

    def setColumnsHidden(self, columns):
        for col in columns:
            self.setColumnHidden(col, True)

    def setupColumns(self, columns, hidden=None):
        for col, opts in columns.items():
            if not hidden:
                self.setColumnHidden(col, opts[1])

            if type(opts[2]) == int:
                self.setColumnWidth(col, opts[2])

            else:
                self.horizontalHeader().setSectionResizeMode(col, opts[2])

        if hidden:
            for col in hidden:
                self.setColumnHidden(int(col), True)

    def setupView(self, view):
        for i, c in enumerate(view):
            if c in ("Device", "Module", "Topic", "FullTopic"):
                self.horizontalHeader().setSectionResizeMode(i, QHeaderView.Stretch)
            else:
                self.horizontalHeader().setSectionResizeMode(i, QHeaderView.ResizeToContents)


class SpinBox(QSpinBox):
    def __init__(self, *args, **kwargs):
        super(SpinBox, self).__init__(*args, **kwargs)
        self.setButtonSymbols(self.NoButtons)
        self.setMinimum(kwargs.get('minimum', 1))
        self.setMaximum(kwargs.get('maximum', 65535))
        self.setSingleStep(kwargs.get('singleStep', 1))
        self.setAlignment(Qt.AlignCenter)


class DoubleSpinBox(QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        super(DoubleSpinBox, self).__init__(*args, **kwargs)
        self.setButtonSymbols(self.NoButtons)
        self.setMinimum(kwargs.get('minimum', 1))
        self.setMaximum(kwargs.get('maximum', 65535))
        self.setDecimals(kwargs.get('precision', 1))
        self.setAlignment(Qt.AlignCenter)


class CheckableAction(QAction):
    def __init__(self, *args, **kwargs):
        super(CheckableAction, self).__init__(*args, **kwargs)
        self.setCheckable(True)


class Toolbar(QToolBar):
    def __init__(self, orientation = Qt.Horizontal, iconsize=32, label_position=Qt.ToolButtonTextUnderIcon, *args, **kwargs):
        super(Toolbar, self).__init__(*args, **kwargs)
        self.setMovable(False)
        self.setIconSize(QSize(iconsize,iconsize))
        self.setOrientation(orientation)
        self.setToolButtonStyle(label_position)

    def addSpacer(self):
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.addWidget(spacer)


class ChannelSlider(QSlider):
    def __init__(self):
        super().__init__()
        self.setOrientation(Qt.Horizontal)
        self.setMinimum(0)
        self.setMaximum(100)
        self.setMaximumWidth(100)
        self.setSingleStep(1)
        self.setPageStep(10)
        self.setTracking(False)


class DictComboBox(QComboBox):
    def __init__(self, items_dict):
        super().__init__()

        for k, v in items_dict.items():
            self.addItem(v, k)


class SliderAction(QWidgetAction):
    def __init__(self, parent, label='', *args, **kwargs):
        super(SliderAction, self).__init__(parent, *args, **kwargs)

        w = QWidget()
        hl = HLayout(5)
        self.slider = ChannelSlider()
        self.slider.setObjectName(label)
        self.value = QLabel("0")
        hl.addWidgets([QLabel(label), self.slider, self.value])
        hl.setStretch(0, 1)
        hl.setStretch(1, 2)
        hl.setStretch(2, 1)
        w.setLayout(hl)
        self.setDefaultWidget(w)

        self.slider.valueChanged.connect(lambda x: self.value.setText(str(x)))


class CmdWikiUrl(QLabel):
    def __init__(self, cmd, title="", *args, **kwargs):
        super(CmdWikiUrl, self).__init__(*args, **kwargs)
        self.setTextFormat(Qt.RichText)
        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.setOpenExternalLinks(True)
        self.setText("<a href={}Commands/#{}>{}</a>".format(docs_url, cmd.lower(), title or cmd))


class HTMLLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super(HTMLLabel, self).__init__(*args, **kwargs)
        self.setTextFormat(Qt.RichText)
        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.setOpenExternalLinks(True)
        self.setWordWrap(True)


class Command(QWidget):
    def __init__(self, command, meta, value=None, *args, **kwargs):
        # print(command, value)
        super(Command, self).__init__(*args, **kwargs)
        self.setMinimumWidth(250)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum))

        vl = VLayout()

        hl = HLayout(0)
        hl.addWidget(QLabel("<b>{}</b>".format(command)))
        hl.addStretch(1)
        hl.addWidget(CmdWikiUrl(command, "Wiki"))

        hl_input = HLayout(0)

        if meta['type'] == "select":
            self.input = QComboBox()
            for k, v in meta['parameters'].items():
                self.input.addItem("{} {}".format(v['description'], "(default)" if v.get("default") else ""), k)

            if meta.get('editable'):
                self.input.setEditable(True)

            if value:
                self.input.setCurrentIndex(value)

        elif meta['type'] == "value":
            self.input = SpinBox(minimum=int(meta['parameters']['min']), maximum=int(meta['parameters']['max']))
            self.input.setMinimumWidth(75)
            if value:
                self.input.setValue(value)
            hl_input.addStretch(1)
            default = meta['parameters'].get('default')
            if default:
                hl_input.addWidget(QLabel("Default: {}".format(default)))
        hl_input.addWidget(self.input)

        vl.addLayout(hl)
        desc = QLabel(meta['description'])
        desc.setWordWrap(True)
        vl.addWidget(desc)
        vl.addLayout(hl_input)

        line = QFrame()
        line.setFrameStyle(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        vl.addWidget(line)
        self.setLayout(vl)


class CommandMultiSelect(QWidget):
    def __init__(self, command, meta, value=None, *args, **kwargs):
        # print(command, value)
        super(CommandMultiSelect, self).__init__(*args, **kwargs)
        self.setMinimumWidth(250)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum))
        self.inputs = []

        vl = VLayout()

        hl = HLayout(0)
        hl.addWidget(QLabel("<b>{}</b>".format(command)))
        hl.addStretch(1)
        hl.addWidget(CmdWikiUrl(command, "Wiki"))
        vl.addLayout(hl)

        desc = QLabel(meta['description'])
        desc.setWordWrap(True)
        vl.addWidget(desc)

        for i, val in enumerate(value):
            cb = QComboBox()
            for k, v in meta['parameters'].items():
                cb.addItem("{}: {} {}".format(k, v['description'], "(default)" if v.get("default") else ""), k)
            cb.setCurrentIndex(val)
            hl_input = HLayout(0)
            hl_input.addWidgets([QLabel("{}: ".format(i+1)), cb])

            self.inputs.append(cb)
            vl.addLayout(hl_input)

        line = QFrame()
        line.setFrameStyle(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        vl.addWidget(line)
        self.setLayout(vl)


class Interlock(QWidget):
    def __init__(self, command, meta, value=None, *args, **kwargs):
        super(Interlock, self).__init__(*args, **kwargs)
        self.setMinimumWidth(250)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum))
        self.input = None
        self.groups = []

        vl = VLayout()

        hl = HLayout(0)
        hl.addWidget(QLabel("<b>{}</b>".format(command)))
        hl.addStretch(1)
        hl.addWidget(CmdWikiUrl(command, "Wiki"))
        vl.addLayout(hl)

        desc = QLabel(meta['description'])
        desc.setWordWrap(True)
        vl.addWidget(desc)

        self.input = QComboBox()

        user_data = ["OFF", "ON"]
        for k, v in meta['parameters'].items():
            self.input.addItem("{} {}".format(v['description'], "(default)" if v.get("default") else ""), user_data[int(k)])

        if value and value.get("Interlock", "OFF") == "OFF":
            self.input.setCurrentIndex(0)
        else:
            self.input.setCurrentIndex(1)
        vl.addWidget(self.input)

        vl_groups = VLayout(0)
        for i in range(4):
            le = QLineEdit()
            le.setAlignment(Qt.AlignCenter)
            group_value = value.get("Groups", [])
            if group_value:
                group_value_list = group_value.split(" ")
                if i < len(group_value_list):
                    group_value = group_value_list[i]
                    le.setText(group_value)
            hl_group = HLayout(0)
            hl_group.addWidgets([QLabel("Group {}".format(i+1)), le])
            vl_groups.addLayout(hl_group)
            self.groups.append(le)
        vl.addLayout(vl_groups)

        line = QFrame()
        line.setFrameStyle(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        vl.addWidget(line)
        self.setLayout(vl)


class PulseTime(QWidget):
    def __init__(self, command, meta, value=None, *args, **kwargs):
        super(PulseTime, self).__init__(*args, **kwargs)
        self.setMinimumWidth(250)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum))
        self.inputs = []

        vl = VLayout()

        hl = HLayout(0)
        hl.addWidget(QLabel("<b>{}</b>".format(command)))
        hl.addStretch(1)
        hl.addWidget(CmdWikiUrl(command, "Wiki"))
        vl.addLayout(hl)

        desc = QLabel(meta['description'])
        desc.setWordWrap(True)
        vl.addWidget(desc)

        vl_groups = VLayout(0)
        for k in sorted(list(value.keys())):
            sb = SpinBox(minimum=int(meta['parameters']['min']), maximum=int(meta['parameters']['max']))
            sb.setValue(value[k])
            hl_group = HLayout(0)
            hl_group.addWidgets([QLabel(k), sb])
            vl_groups.addLayout(hl_group)
            self.inputs.append(sb)
        vl.addLayout(vl_groups)

        line = QFrame()
        line.setFrameStyle(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        vl.addWidget(line)
        self.setLayout(vl)



from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QDockWidget, QTreeWidget, QTreeWidgetItem

# TODO: merge get_top_item() and get_nested_item() to a recursive generator

item_icons = {
    "Time": "time.png",
}

item_units = {
    # ENERGY units
    "Voltage": "V",
    "Current": "A",
    "Power": "W",
    "ApparentPower": "VA",
    "ReactivePower": "VAR",
    "Frequency": "Hz",
    "Total": "kWh",
    "Yesterday": "kWh",
    "Today": "kWh",
}

class TelemetryWidget(QDockWidget):
    def __init__(self, device, *args, **kwargs):
        super(TelemetryWidget, self).__init__(*args, **kwargs)
        self.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        self.setWindowTitle(device.name)

        self.tree_items = {}

        self.tree = QTreeWidget()
        self.setWidget(self.tree)
        self.tree.setColumnCount(2)
        self.tree.setHeaderHidden(True)

        device.update_telemetry.connect(self.update_telemetry)

        self.device = device

    def get_top_item(self, name):
        item = self.tree_items.get(name)
        if not item:
            item = QTreeWidgetItem()
            item.setText(0, name)
            item.setIcon(0, QIcon("GUI/icons/{}".format(item_icons.get(name, ""))))
            self.tree.addTopLevelItem(item)
            self.tree_items[name] = item
        return item

    def get_nested_item(self, parent, name):
        nested_items = self.tree_items.get(parent)
        if nested_items:
            _top_item = nested_items["_top_item"]
            item = nested_items.get(name)
            if not item:
                item = QTreeWidgetItem()
                item.setText(0, name)
                item.setIcon(0, QIcon("GUI/icons/{}".format(item_icons.get(name, ""))))
                _top_item.addChild(item)
                nested_items[name] = item
            return item
        else:
            _top_item = QTreeWidgetItem()
            _top_item.setText(0, parent)
            self.tree.addTopLevelItem(_top_item)
            self.tree_items[parent] = {"_top_item": _top_item}

            item = QTreeWidgetItem()
            item.setText(0, name)
            _top_item.addChild(item)
            self.tree_items[parent][name] = item
            return item


    @pyqtSlot()
    def update_telemetry(self):
        t = self.device.t

        time = t.get("Time")

        time_item = self.get_top_item("Time")
        time_item.setText(1, time.replace("T", " "))

        for key in sorted(t.keys()):
            if key != "Time":
                v = t[key]
                if isinstance(v, dict):
                    for nested_key, nested_v in v.items():
                        nested_item = self.get_nested_item(key, nested_key)
                        nested_item.setText(1, "{} {}".format(nested_v, item_units.get(nested_key, "")))

                else:
                    item = self.get_top_item(key)
                    item.setText(1, "{}".format(v))

        self.tree.expandAll()
        self.tree.resizeColumnToContents(0)
        self.tree.resizeColumnToContents(1)



commands = {
  "BlinkCount": {
        "description": "Number of relay toggles (blinks)",
        "type": "value",
        "parameters": {
            "min": "0",
            "max": "32000",
            "default": "10"
        }
    },

  "BlinkTime": {
        "description": "Blink duration [0.1s]",
        "type": "value",
        "parameters": {
            "min": "2",
            "max": "3600",
            "default": "10"
        }
    },

  "ButtonDebounce": {
        "description": "Button debounce timing [ms]",
        "type": "value",
        "parameters": {
            "min": "40",
            "max": "100",
            "default": "50"
        }
    },

  "ButtonRetain": {
        "description": "Add MQTT retain flag on button press",
        "type": "select",
        "parameters": {
            "0": {"description": "Disabled", "default": "True"},
            "1": {"description": "Enabled"}
        }
    },

  "ButtonTopic": {
        "description": "MQTT button topic",
        "type": "select",
        "editable": "True",
        "parameters": {
            "0": {"description": "Disable use of ButtonTopic", "default": "True"},
            "1": {"description": "Set ButtonTopic to device %topic%"},
            "2": {"description": "Reset ButtonTopic to firmware default"}
        }
    },

  "Interlock": {
        "description": "Relay interlock mode",
        "type": "select",
        "parameters": {
            "0": {"description": "Self-locking mode for all relays", "default": "True"},
            "1": {"description": "Interlock mode for selected relays"}
        }
    },

  "PowerOnState": {
        "description": "Control relay state when the device is powered up",
        "type": "select",
        "parameters": {
            "0": {"description": "Keep relay(s) OFF after power up"},
            "1": {"description": "Turn relay(s) ON after power up"},
            "2": {"description": "Toggle relay(s) from last saved state"},
            "3": {"description": "Switch relay(s) to their last saved state", "default": "True"},
            "4": {"description": "Turn relay(s) ON and disable further relay control"},
            "5": {"description": "Turn relay(s) ON after a PulseTime period"}
        }
    },

  "PowerRetain": {
        "description": "Add MQTT power retain flag on status update",
        "type": "select",
        "parameters": {
            "0": {"description": "Disabled", "default": "True"},
            "1": {"description": "Enabled"}
        }
    },

  "PulseTime": {
        "description": "Control the relay PulseTime",
        "type": "value",
        "parameters": {
            "min": "0",
            "max": "64900"
        }
    },

  "SwitchMode": {
        "description": "Configure the behavior of a physical input peripheral configured as a Switch<x>",
        "type": "select",
        "parameters": {
            "0": {"description": "Toggle", "default": "True"},
            "1": {"description": "Follow [0=Off, 1=On]"},
            "2": {"description": "Inverted follow [0=On, 1=Off]"},
            "3": {"description": "Pushbutton [Default 0, 1=Toggle]"},
            "4": {"description": "Inverted pushbutton [Default 1, 0=Toggle]"},
            "5": {"description": "Pushbutton with hold [Default 1, 0=Toggle, 2=Hold]"},
            "6": {"description": "Inverted pushbutton with hold [Default 0, 1=Toggle, 2=Hold]"},
            "7": {"description": "Pushbutton toggle [0=Toggle, 1=Toggle]"}
        }
    },

  "SwitchDebounce": {
        "description": "Switch debounce timing [ms]",
        "type": "value",
        "parameters": {
            "min": "40",
            "max": "100",
            "default": "50"
        }
    },

  "SwitchRetain": {
        "description": "Add MQTT retain flag on switch press",
        "type": "select",
        "parameters": {
            "0": {"description": "Disabled", "default": "True"},
            "1": {"description": "Enabled"}
        }
    }
}